# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)


class Database:
    def lookup_parts(self):
        raise NotImplementedError("")

    def lookup_dicts(self):
        raise NotImplementedError("")

    def count(self, request):
        raise NotImplementedError("")

    def load(self, iterator):
        raise NotImplementedError("")

    def sel(self, selection):
        raise NotImplementedError("")

    def order_by(self, order):
        raise NotImplementedError("")



# (C) Copyright 2022 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import logging
import os
from abc import abstractmethod
from collections import namedtuple
from typing import List
from urllib.parse import urljoin

import requests
from multiurl import robust

from climetlab.core.caching import auxiliary_cache_file, cache_file
from climetlab.core.index import (
    Index,
    MaskIndex,
    MultiIndex,
    Order,
    OrderOrSelection,
    Selection,
)
from climetlab.decorators import cached_method
from climetlab.indexing.database.json import JsonDatabase
from climetlab.indexing.database.sql import SqlDatabase
from climetlab.readers.grib.codes import GribField, get_messages_positions
from climetlab.readers.grib.fieldset import FieldSetMixin
from climetlab.utils import progress_bar
from climetlab.utils.parts import Part
from climetlab.utils.serialise import register_serialisation

LOG = logging.getLogger(__name__)


class FieldSet(FieldSetMixin, Index):
    _availability = None

    def __init__(self, *args, **kwargs):

        Index.__init__(self, *args, **kwargs)

    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskFieldSet(*args, **kwargs)

    @property
    def availability(self):
        raise NotImplementedError()
        if self._availability is not None:
            return self._availability.tree()
        LOG.debug("Building availability")

        def dicts():
            for i in self:
                dic = i.to_dict()  # GribField.to_dict() is not implemented
                i = {k: v for k, v in dic.items() if not k.startswith("_")}
                yield i

        from climetlab.utils.availability import Availability

        self._availability = Availability(dicts())
        return self.availability


class MaskFieldSet(FieldSet, MaskIndex):
    def __init__(self, *args, **kwargs):
        MaskIndex.__init__(self, *args, **kwargs)


class MultiFieldSet(FieldSet, MultiIndex):
    def __init__(self, *args, **kwargs):
        MultiIndex.__init__(self, *args, **kwargs)


class FieldSetInFiles(FieldSet):
    # Remote Fieldsets (with urls) are also here,
    # as the actual fieldset is accessed on a file in cache.
    # This class changes the interface (_getitem__ and __len__)
    # into the interface (part and number_of_parts).
    def __getitem__(self, n):
        part = self.part(n)
        return GribField(part.path, part.offset, part.length)

    def __len__(self):
        return self.number_of_parts()

    @abstractmethod
    def part(self, n):
        self._not_implemented()

    @abstractmethod
    def number_of_parts(self):
        self._not_implemented()


class FieldsetInFilesWithDBIndex(FieldSetInFiles):
    def __init__(self, db, **kwargs):
        """Should not be instanciated directly.
        The public API are the constructors "_from*()" class methods.
        """

        self.db = db

        # self._cache is a tuple : (first, length, result). It holds one chunk of the db.
        # The third element (result) is a list of size length.
        self._cache = None

        super().__init__(**kwargs)

    @classmethod
    def from_iterator(
        cls,
        iterator,
        cache_metadata,
        selection=None,
        order_by=None,
        **kwargs,
    ):
        def load(target, *args):
            LOG.debug(f"Building db in {target}")
            db = cls.DBCLASS(target)
            db.load(iterator)

        db_name = cache_file(
            "grib-index",
            load,
            cache_metadata,
            hash_extra=cls.DBCLASS.VERSION,
            extension=cls.DBCLASS.EXTENSION,
        )

        db = cls.DBCLASS(db_name)

        new = cls(db=db, **kwargs)
        new = new.sel(selection)
        new = new.order_by(order_by)
        return new

    @classmethod
    def from_url(cls, url, patch_entry=None, **kwargs):
        """Create a database from a given url"""

        # If this is a file, open it without download
        if os.path.exists(url):
            return cls.from_file(path=url, **kwargs)
        if url.startswith("file://") and os.path.exists(url[7:]):
            return cls.from_file(path=url[7:], **kwargs)

        # Request to download the data
        r = robust(requests.get)(url, stream=True)
        r.raise_for_status()
        try:
            size = int(r.headers.get("Content-Length"))
        except Exception:
            size = None

        if patch_entry is None:

            def absolute_url(entry):  # closure on "url"
                entry["_path"] = urljoin(url, entry.pop("_path"))
                return entry

            patch_entry = absolute_url

        def progress(iterable):
            pbar = progress_bar(iterable=iterable, desc="Downloading index", total=size)
            for line in pbar:
                yield line
                pbar.update(len(line) + 1)

        def parse_lines(lines):
            for line in lines:
                entry = json.loads(line)
                entry = patch_entry(entry)
                yield entry

        iterator = r.iter_lines()
        iterator = progress(iterator)
        iterator = parse_lines(iterator)

        return cls.from_iterator(
            iterator=iterator,
            cache_metadata={"url": url},
            **kwargs,
        )

    @classmethod
    def from_file(cls, path, **kwargs):
        directory = os.path.dirname(path)
        size = os.path.getsize(path)

        def progress(lines):
            pbar = progress_bar(iterable=lines, desc="Parsing index", total=size)
            for line in pbar:
                yield line
                pbar.update(len(line) + 1)

        def parse_lines(lines):
            for line in lines:
                entry = json.loads(line)
                if not os.path.isabs(entry["_path"]):
                    entry["_path"] = os.path.join(directory, entry["_path"])
                yield entry

        iterator = open(path)
        iterator = progress(iterator)
        iterator = parse_lines(iterator)

        return cls.from_iterator(
            iterator=iterator,
            cache_metadata={"path": path},
            **kwargs,
        )

    @classmethod
    def from_existing_db(cls, db_path, **kwargs):
        assert os.path.exists(db_path)
        return cls(cls.DBCLASS(db_path), **kwargs)

    @property
    def availability(self):
        if self._availability is not None:
            return self._availability.tree()

        LOG.debug("Building availability")

        def dicts():
            for i in self.db.lookup_dicts():
                dic = {}
                for k, v in i.items():
                    if k.startswith("_"):
                        continue
                    if k.startswith("i_"):
                        k = k[2:]
                    dic[k] = v
                yield dic

        from climetlab.utils.availability import Availability

        self._availability = Availability(dicts())
        return self.availability


class FieldsetInFilesWithJsonIndex(FieldsetInFilesWithDBIndex):
    DBCLASS = JsonDatabase

    @cached_method
    def _lookup_parts(self):
        return self.db.lookup_parts()

    def part(self, n):
        return self._lookup_parts()[n]

    def number_of_parts(self):
        return len(self._lookup_parts())


SqlResultCache = namedtuple("SqlResultCache", ["first", "length", "result"])


class FieldsetInFilesWithSqlIndex(FieldsetInFilesWithDBIndex):

    DBCLASS = SqlDatabase
    DB_CACHE_SIZE = 50_000

    def __init__(self, *args, _filters=None, **kwargs):
        """
        _filters are used to keep the state of the db
        It is a list of **already applied** filters, not a list of filter to apply.
        Use the method apply_filters for this.
        """

        # _filter
        self._filters = _filters or []
        super().__init__(*args, **kwargs)

    def apply_filters(self, filters: List[OrderOrSelection]):
        obj = self
        for f in filters:
            obj = obj.filter(f)
        return obj

    def filter(self, filter):
        if filter.is_empty:
            return self

        db = self.db.filter(filter)
        _filters = self._filters + [filter]
        return self.__class__(db=db, _filters=_filters)

    def sel(self, *args, **kwargs):
        return self.filter(Selection(*args, **kwargs))

    def order_by(self, *args, **kwargs):
        return self.filter(Order(*args, **kwargs))

    def part(self, n):
        if self._cache is None or not (
            self._cache.first <= n < self._cache.first + self._cache.length
        ):
            first = n // self.DB_CACHE_SIZE
            result = self.db.lookup_parts(limit=self.DB_CACHE_SIZE, offset=first)
            self._cache = SqlResultCache(first, len(result), result)
        return self._cache.result[n % self.DB_CACHE_SIZE]

    @cached_method
    def number_of_parts(self):
        return self.db.count()


register_serialisation(
    FieldsetInFilesWithSqlIndex,
    lambda x: [x.db.db_path, x._filters],
    lambda x: FieldsetInFilesWithSqlIndex(db=SqlDatabase(x[0])).apply_filters(
        filters=x[1]
    ),
)


class FieldSetInOneFile(FieldSetInFiles):
    VERSION = 1

    def __init__(self, path, **kwargs):
        assert isinstance(path, str), path

        self.path = path
        self.offsets = None
        self.lengths = None
        self.mappings_cache_file = auxiliary_cache_file(
            "grib-index",
            path,
            content="null",
            extension=".json",
        )

        if not self._load_cache():
            self._build_offsets_lengths_mapping()

        super().__init__(**kwargs)

    def _build_offsets_lengths_mapping(self):

        offsets = []
        lengths = []

        for offset, length in get_messages_positions(self.path):
            offsets.append(offset)
            lengths.append(length)

        self.offsets = offsets
        self.lengths = lengths

        self._save_cache()

    def _save_cache(self):
        try:
            with open(self.mappings_cache_file, "w") as f:
                json.dump(
                    dict(
                        version=self.VERSION,
                        offsets=self.offsets,
                        lengths=self.lengths,
                    ),
                    f,
                )
        except Exception:
            LOG.exception("Write to cache failed %s", self.mappings_cache_file)

    def _load_cache(self):
        try:
            with open(self.mappings_cache_file) as f:
                c = json.load(f)
                if not isinstance(c, dict):
                    return False

                assert c["version"] == self.VERSION
                self.offsets = c["offsets"]
                self.lengths = c["lengths"]
                return True
        except Exception:
            LOG.exception("Load from cache failed %s", self.mappings_cache_file)

        return False

    def part(self, n):
        return Part(self.path, self.offsets[n], self.lengths[n])

    def number_of_parts(self):
        return len(self.offsets)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import json
import logging
import os

from climetlab.utils.parts import Part

from . import Database

LOG = logging.getLogger(__name__)


class JsonDatabase(Database):
    VERSION = 1
    EXTENSION = ".json"

    def __init__(self, db_path):
        self.entries = []
        self.db_path = db_path
        if os.path.exists(db_path):
            with open(db_path) as f:
                for entry in f:
                    self.entries.append(json.loads(entry))

    def load(self, iterator):
        self.entries = list(iterator)

        with open(self.db_path, "w") as f:
            for entry in self.entries:
                print(json.dumps(entry), file=f)

        return len(self.entries)

    def lookup_parts(self, order=None, **kwargs):
        parts = []
        for e in self.entries:
            parts.append(Part(e["_path"], e["_offset"], e["_length"]))
        return Part.resolve(parts, os.path.dirname(self.db_path))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

import numpy as np

LOG = logging.getLogger(__name__)


class PytorchMixIn:
    def to_pytorch(self, offset, data_loader_kwargs=None):
        import torch

        # sometimes (!) causes an Exception:
        # gribapi.errors.UnsupportedEditionError: Edition not supported.
        num_workers = 1

        out = self._to_pytorch_wrapper_class()(self, offset)

        DATA_LOADER_KWARGS_DEFAULT = dict(
            batch_size=128,
            # multi-process data loading
            # use as many workers as you have cores on your machine
            num_workers=num_workers,
            # default: no shuffle, so need to explicitly set it here
            shuffle=True,
            # uses pinned memory to speed up CPU-to-GPU data transfers
            # see https://pytorch.org/docs/stable/notes/cuda.html#cuda-memory-pinning
            pin_memory=True,
            # function used to collate samples into batches
            # if None then Pytorch uses the default collate_fn (see below)
            collate_fn=None,
        )
        data_loader_kwargs_ = {k: v for k, v in DATA_LOADER_KWARGS_DEFAULT.items()}
        if data_loader_kwargs:
            data_loader_kwargs_.update(data_loader_kwargs)

        return torch.utils.data.DataLoader(out, **data_loader_kwargs_)

    def _to_pytorch_wrapper_class(self):
        import torch

        class WrapperWeatherBenchDataset(torch.utils.data.Dataset):
            def __init__(self, ds, offset) -> None:
                super().__init__()
                self.ds = ds
                self.stats = ds.statistics()
                self.offset = offset

            def __len__(self):
                """Returns the length of the dataset. This is important! Pytorch must know this."""
                return len(self.ds) - self.offset
                return self.ds.stats["count"] - self.ds.offset

            def __getitem__(self, i):  # -> Tuple[np.ndarray, ...]:
                """Returns the i-th sample (x, y). Pytorch will take care of the shuffling after each epoch."""
                # Q: if self is a iterator, would ds[i] read the data from 0 to i, just to provide i?

                x, Y = (
                    self.ds[i].to_numpy()[None, ...],
                    self.ds[i + self.offset].to_numpy()[None, ...],
                )
                x = x.astype(np.float32)
                Y = Y.astype(np.float32)
                return x, Y

        return WrapperWeatherBenchDataset



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import hashlib
import json
import logging
from abc import abstractmethod

from climetlab.decorators import alias_argument
from climetlab.sources import Source

LOG = logging.getLogger(__name__)


class OrderOrSelection:
    def __init__(self, *args, **kwargs):
        """Parse args and kwargs to build a dictionary in self.order"""
        self.dic = {}

        for arg in args:
            if isinstance(arg, dict):
                arg = self.normalize_naming(**arg)
            res_ok = self.parse_arg(arg)
            if res_ok is False:
                raise ValueError(f"Invalid argument of type({type(arg)}): {arg}")

        kwargs = self.normalize_naming(**kwargs)
        for k, v in kwargs.items():
            self.parse_kwarg(k, v)

    @alias_argument("levelist", ["level"])
    @alias_argument("param", ["variable", "parameter"])
    @alias_argument("number", ["realization", "realisation"])
    @alias_argument("class", "klass")
    def normalize_naming(self, **kwargs):
        return kwargs

    def parse_arg(self, arg):
        # Returns True of argument has been parsed.
        if arg is None:
            return True
        if isinstance(arg, dict):
            for k, v in arg.items():
                self.parse_kwarg(k, v)
            return True
        return False

    def __str__(self):
        return f"{self.__class__.__name__}({self.dic})"

    @property
    def is_empty(self):
        if not self.dic:
            return True
        return False

    def h(self, *args, **kwargs):
        m = hashlib.sha256()
        m.update(json.dumps(args, sort_keys=True).encode("utf-8"))
        m.update(json.dumps(kwargs, sort_keys=True).encode("utf-8"))
        m.update(json.dumps(self.dic, sort_keys=True).encode("utf-8"))
        return m.hexdigest()


class Selection(OrderOrSelection):
    @property
    def selection(self):
        return self.dic

    def parse_kwarg(self, k, v):
        if v is not None and not isinstance(v, (list, tuple)):
            v = [v]
        if isinstance(v, (list, tuple)):
            v = [str(_) for _ in v]
        self.dic[k] = v
        return

    def match_element(self, element):
        for k, v in self.dic.items():
            if v is None:
                continue
            value = element.metadata(k)
            # value = grib_naming({key:value})[key]
            if isinstance(v, (list, tuple)):
                if value in v:
                    continue
                if str(value) in v:
                    continue
                return False
            if value == v:
                continue
            if str(value) == v:
                continue
            return False
        return True


class Order(OrderOrSelection):
    def __init__(self, *args, **kwargs):
        self._rankers = None
        super().__init__(*args, **kwargs)

    @property
    def order(self):
        return self.dic

    def items(self):
        if self.is_empty:
            return

        for k, v in self.dic.items():
            yield k, v

        from climetlab.indexing.database.sql import GRIB_INDEX_KEYS

        for k in [_ for _ in GRIB_INDEX_KEYS if _ not in self.dic]:
            yield k, self[k]

    def __getitem__(self, key):
        # Default is ascending order
        return self.dic.get(key, "ascending")

    def parse_arg(self, arg):
        if super().parse_arg(arg):
            return True
        if isinstance(arg, str):
            self.dic[arg] = "ascending"
            return True
        return False

    def parse_kwarg(self, k, v):
        if isinstance(v, (list, tuple)):
            v = [str(_) for _ in v]  # processing only strings from now.
        if (v == "ascending") or (v == "descending") or isinstance(v, (list, tuple)):
            self.dic[k] = v
            return

        if v is None:
            self.dic[k] = "ascending"
            return

        self.dic[k] = "ascending"
        return
        # raise ValueError(f"Invalid argument for {k}: {v} ({type(v)})")

    def build_rankers(self):
        if self._rankers is not None:
            return self._rankers

        from climetlab.indexing.database.sql import GRIB_INDEX_KEYS

        keys = [_ for _ in self.dic.keys()]
        keys += [_ for _ in GRIB_INDEX_KEYS if _ not in keys]

        key_types = {}
        dict_of_dicts = dict()

        for key in keys:
            lst = self.dic.get(key, None)

            if isinstance(lst, (tuple, list)):
                dict_of_dicts[key] = dict(zip(lst, range(len(lst))))
                key_types[key] = "explicit"
                continue

            if lst == "ascending" or lst is None:
                dict_of_dicts[key] = lambda value: value
                key_types[key] = "ascending"
                continue

            raise ValueError(f"Invalid argument {lst}")

        self._rankers = keys, key_types, dict_of_dicts
        return self._rankers

    def rank(self, element):
        keys, key_types, dict_of_dicts = self.build_rankers()
        ranks = []
        for k in keys:
            value = element.metadata(k)

            if key_types[k] == "ascending":
                ranks.append(value)
                continue

            if key_types[k] == "explicit":
                value = str(value)
                ranks.append(dict_of_dicts[k][value])
                continue

            assert False, (k, key_types[k], element)

        return tuple(ranks)


class Index(Source):
    @classmethod
    def new_mask_index(self, *args, **kwargs):
        return MaskIndex(*args, **kwargs)

    def __init__(self, *args, order_by=None, **kwargs):
        self._init_args = args
        self._init_kwargs = kwargs
        self._init_order_by = order_by

    def mutate(self):
        source = self
        source = source.sel(*self._init_args, **self._init_kwargs)
        source = source.order_by(*self._init_args, **self._init_kwargs)
        source = source.order_by(self._init_order_by)
        return source

    @abstractmethod
    def __getitem__(self, n):
        self._not_implemented()

    @abstractmethod
    def __len__(self):
        self._not_implemented()

    @abstractmethod
    def sel(self, *args, **kwargs):
        """Filter elements on their metadata(), according to kwargs.
        Returns a new index object.
        """
        selection = Selection(*args, **kwargs)
        if selection.is_empty:
            return self

        indices = []
        for i, element in enumerate(self):
            if selection.match_element(element):
                indices.append(i)

        return self.new_mask_index(self, indices)

    def order_by(self, *args, **kwargs):
        """Default order_by method.
        It expects that calling self[i] returns an element that and Order object can rank
        (i.e. order.rank(element) -> tuple).
        then it sorts the elements according to the tuples.

        Returns a new index object.
        """

        order = Order(*args, **kwargs)
        if order.is_empty:
            return self

        for k, v in order.order.items():
            assert isinstance(v, (tuple, list)) or v in [
                "ascending",
                None,
            ], f"Unsupported order: {v}, Supported values: 'ascending'."

        class Sorter:
            def __init__(_self, index, order):
                """Uses the order to sort the index"""

                _self.index = index
                _self.order = order
                _self._cache = [None] * len(_self.index)

            def __call__(_self, i):
                if _self._cache[i] is None:
                    element = _self.index[i]
                    _self._cache[i] = _self.order.rank(element)
                return _self._cache[i]

        sorter = Sorter(self, order)

        result = list(range(len(self)))
        indices = sorted(result, key=sorter)
        return self.new_mask_index(self, indices)


class MaskIndex(Index):
    def __init__(self, index, indices):
        self.index = index
        self.indices = indices
        super().__init__(
            *self.index._init_args,
            order_by=self.index._init_order_by,
            **self.index._init_kwargs,
        )

    def __getitem__(self, n):
        n = self.indices[n]
        return self.index[n]

    def __len__(self):
        return len(self.indices)


class MultiIndex(Index):
    def __init__(self, indexes, *args, **kwargs):
        self.indexes = list(indexes)
        super().__init__(*args, **kwargs)
        # self.indexes = list(i for i in indexes if len(i))
        # TODO: propagate  index._init_args, index._init_order_by, index._init_kwargs, for each i in indexes?

    def sel(self, *args, **kwargs):
        selection = Selection(*args, **kwargs)
        if selection.is_empty:
            return self
        return self.__class__(i.sel(*args, **kwargs) for i in self.indexes)

    def __getitem__(self, n):
        k = 0
        while n >= len(self.indexes[k]):
            n -= len(self.indexes[k])
            k += 1
        return self.indexes[k][n]

    def __len__(self):
        return sum(len(i) for i in self.indexes)

    def graph(self, depth=0):
        print(" " * depth, self.__class__.__name__)
        for s in self.indexes:
            s.graph(depth + 3)


class ForwardingIndex(Index):
    def __init__(self, index):
        self.index = index

    def __len__(self):
        return len(self.index)


class ScaledField:
    def __init__(self, field, offset, scaling):
        self.field = field
        self.offset = offset
        self.scaling = scaling

    def to_numpy(self):
        return (self.field.to_numpy() - self.offset) * self.scaling


class ScaledIndex(ForwardingIndex):
    def __init__(self, index, offset, scaling):
        super().__init__(index)
        self.offset = offset
        self.scaling = scaling

    def __getitem__(self, n):
        return ScaledField(self.index[n], self.offset, self.scaling)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import logging
import os
import sqlite3

import climetlab as cml
from climetlab.core.index import Order, OrderOrSelection, Selection
from climetlab.utils import tqdm
from climetlab.utils.parts import Part

from . import Database

LOG = logging.getLogger(__name__)


GRIB_INDEX_KEYS = [
    "class",
    "stream",
    "levtype",
    "type",
    "expver",
    "date",
    "hdate",
    "andate",
    "time",
    "antime",
    "reference",
    "step",
    "anoffset",
    "verify",
    "fcmonth",
    "fcperiod",
    "leadtime",
    "opttime",
    "origin",
    "domain",
    "method",
    "diagnostic",
    "iteration",
    "number",
    "quantile",
    "levelist",
    # "latitude",  # in the MARS vocabulary but not used.
    # "longitude",  # in the MARS vocabulary but not used.
    "range",
    "param",
    "ident",
    "obstype",
    "instrument",
    "reportype",
    "frequency",  # for 2-d wave-spectra products
    "direction",  # for 2-d wave-spectra products
    "channel",  # for ea, ef
]


class SqlSorter:
    @property
    def _func_name(self):
        return f"userorder_{self.view}"

    def __init__(self, order, view):
        self.order = order
        self.view = view

        self.dict_of_dicts = dict()
        self.order_lst = []

        # TODO: To improve speed, we could use ASC or DESC when lst is already sorted
        # TODO: move GRIB_INDEX_KEYS and two comments above to upper class
        # Use mars keys order by default
        # But make sure the order provided by the user
        # in the order override this default order.

        if order is None or order.is_empty:
            return

        for key, lst in self.order.items():
            self._add_key(key, lst)

    def _add_key(self, key, lst):
        if lst is None:
            self.order_lst.append(f"i_{key}")
            return
        if lst == "ascending":
            self.order_lst.append(f"i_{key} ASC")
            return
        if lst == "descending":
            self.order_lst.append(f"i_{key} DESC")
            return
        if not isinstance(lst, (list, tuple)):
            lst = [lst]

        lst = [str(_) for _ in lst]  # processing only strings from now.

        self.dict_of_dicts[key] = dict(zip(lst, range(len(lst))))
        self.order_lst.append(f'{self._func_name}("{key}",i_{key})')

    @property
    def order_statement(self):
        if not self.order_lst:
            assert not self.dict_of_dicts, self.dict_of_dicts
            return ""
        return "ORDER BY " + ",".join(self.order_lst)

    def create_sql_function_if_needed(self, connection):
        if not self.dict_of_dicts:
            return

        dict_of_dicts = self.dict_of_dicts  # avoid creating closure on self.

        def order_func(k, v):
            return dict_of_dicts[k][v]

        connection.create_function(self._func_name, 2, order_func)


class SqlDatabase(Database):
    VERSION = 3
    EXTENSION = ".db"

    def __init__(self, db_path, view="entries", _connection=None):

        self.view = view
        self.db_path = db_path
        self._connection = _connection

        LOG.debug("DB %s %s", self.db_path, self.view)

    @property
    def connection(self):
        if self._connection is None:
            LOG.debug(f"Connecting to DB ({self.db_path})")
            self._connection = sqlite3.connect(self.db_path)
        return self._connection

    def load(self, iterator):
        with self.connection as conn:

            # The i_ is to avoid clashes with SQL keywords
            i_columns = [f"i_{n}" for n in GRIB_INDEX_KEYS]
            columns_defs = ",".join([f"{c} TEXT" for c in i_columns])
            create_statement = f"""CREATE TABLE IF NOT EXISTS entries (
                path    TEXT,
                offset  INTEGER,
                length  INTEGER,
                {columns_defs}
                );"""
            LOG.debug("%s", create_statement)
            conn.execute(create_statement)

            columns = ",".join(i_columns)
            values = ",".join(["?"] * (3 + len(i_columns)))
            insert_statement = f"""
            INSERT INTO entries (path, offset, length, {columns})
            VALUES({values});
            """
            LOG.debug("%s", insert_statement)

            count = 0

            for entry in iterator:
                assert isinstance(entry, dict), (type(entry), entry)
                values = [entry["_path"], entry["_offset"], entry["_length"]] + [
                    entry.get(n) for n in GRIB_INDEX_KEYS
                ]
                conn.execute(insert_statement, tuple(values))
                count += 1

            assert count >= 1
            LOG.info("Added %d entries", count)

            # connection.execute(f"CREATE INDEX path_index ON entries (path);")
            pbar = tqdm(i_columns + ["path"], desc="Building indexes")
            for n in pbar:
                pbar.set_description(f"Building index for {n}")
                conn.execute(f"CREATE INDEX IF NOT EXISTS {n}_index ON entries ({n});")

            return count

    def _conditions(self, selection):
        if selection is None or selection.is_empty:
            return ""
        conditions = []
        for k, b in selection.dic.items():
            if b is None or b == cml.ALL:
                continue
            elif isinstance(b, (list, tuple)):
                if len(b) == 1:
                    conditions.append(f"i_{k}='{b[0]}'")
                    continue
                w = ",".join([f"'{x}'" for x in b])
                conditions.append(f"i_{k} IN ({w})")
            else:
                conditions.append(f"i_{k}='{b}'")

        if not conditions:
            return ""
        return " WHERE " + " AND ".join(conditions)

    def _columns_names_without_i_(self):
        cursor = self.connection.execute("PRAGMA table_info(entries)")
        out = []
        for x in cursor.fetchall():
            name = x[1]
            if name.startswith("i_"):
                name = name[2:]  # remove "i_"
                out.append(name)
        return out

    def filter(self, filter: OrderOrSelection):
        view = self.view
        view += "_" + filter.h(parent_view=self.view)

        if isinstance(filter, Selection):
            order = None
            selection = filter
        elif isinstance(filter, Order):
            selection = None
            order = filter
        else:
            assert False, (type(filter), filter)

        connection = self.connection

        conditions_statement = self._conditions(selection)
        sorter = SqlSorter(order, view)
        statement = (
            f"CREATE TEMP VIEW IF NOT EXISTS {view} AS SELECT * "
            + f"FROM {self.view} {conditions_statement} {sorter.order_statement};"
        )

        sorter.create_sql_function_if_needed(connection)
        LOG.debug("%s", statement)
        for i in connection.execute(statement):
            LOG.error(str(i))

        return self.__class__(
            self.db_path,
            view=view,
            _connection=self._connection,
        )

    def lookup_parts(self, limit=None, offset=None, resolve_paths=True):
        """
        Look into the database and provide entries as Parts.
        limit: Returns only "limit" entries (used for paging).
        offset: Skip the first "offset" entries (used for paging).
        """

        _names = ["path", "offset", "length"]
        parts = []
        for path, offset, length in self._execute_select(_names, limit, offset):
            parts.append(Part(path, offset, length))
        if resolve_paths:
            parts = Part.resolve(parts, os.path.dirname(self.db_path))
        return parts

    def lookup_dicts(self, keys=None, limit=None, offset=None):
        """
        From a list of keys, return dicts with these columns of the database.
        limit: Returns only "limit" entries (used for paging).
        offset: Skip the first "offset" entries (used for paging).
        """
        if keys is None:
            keys = self._columns_names_without_i_()
        assert isinstance(keys, (list, tuple)), keys

        _names = [f"i_{x}" for x in keys]
        dicts = []
        for tupl in self._execute_select(_names, limit, offset):
            dic = {k: v for k, v in zip(_names, tupl)}
            dicts.append(dic)
        return dicts

    def _execute_select(self, names, limit, offset):
        names_str = ",".join([x for x in names]) if names else "*"
        limit_str = f" LIMIT {limit}" if limit is not None else ""
        offset_str = f" OFFSET {offset}" if offset is not None else ""

        statement = f"SELECT {names_str} FROM {self.view} {limit_str} {offset_str};"
        LOG.debug("%s", statement)

        for tupl in self.connection.execute(statement):
            yield tupl

    def count(self):
        statement = f"SELECT COUNT(*) FROM {self.view};"
        for result in self.connection.execute(statement):
            return result[0]
        assert False, statement  # Fail if result is empty.

    def _dump_dicts(self):
        names = self._columns_names_without_i_()
        names_str = ",".join([f"i_{x}" for x in names])
        statement = f"SELECT path,offset,length,{names_str} FROM entries ;"
        for tupl in self.connection.execute(statement):
            yield {k: v for k, v in zip(["_path", "_offset", "_length"] + names, tupl)}

    def dump_dicts(self, remove_none=True):
        for dic in self._dump_dicts():
            if remove_none:
                dic = {k: v for k, v in dic.items() if v is not None}
            yield dic

    def duplicate_db(self, filename, **kwargs):
        new = SqlDatabase(db_path=filename)
        new.load(self.dump_dicts(**kwargs))



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging

LOG = logging.getLogger(__name__)


def to_tfdataset(
    features,
    targets=None,
    total_size=None,
    num_parallel_calls=10,
    prefetch=1024,
    **kwargs,
):

    import tensorflow as tf

    if features is not None and not callable(features):
        if total_size is None:
            LOG.debug("No total_size specified, infering from features.")
            total_size = len(features)

        _features = features

        def features(i):
            return _features[i].to_numpy()

    if targets is not None and not callable(targets):
        _targets = targets

        def targets(i):
            return _targets[i].to_numpy()

    assert total_size is not None

    def map_fn(i):
        i = int(i)
        return features(i)

    @tf.function
    def tf_map_fn(i):
        return tf.py_function(func=map_fn, inp=[i], Tout=tf.float32)

    def map_label_fn(i):
        i = int(i)
        return targets(i)

    @tf.function
    def tf_map_label_fn(i):
        return tf.py_function(func=map_label_fn, inp=[i], Tout=tf.float32)

    def dataset(mapping):
        return (
            tf.data.Dataset.range(total_size)
            .map(mapping, num_parallel_calls=num_parallel_calls)
            .prefetch(prefetch)
        )

    if targets is None:
        return dataset(tf_map_fn)

    return tf.data.Dataset.zip(
        (
            dataset(tf_map_fn),
            dataset(tf_map_label_fn),
        )
    )


class TensorflowMixIn:
    def to_tfdataset(
        self,
        labels=None,
        targets=None,
        **kwargs,
    ):
        if targets is None:  # rename "labels" into "targets" ?
            targets = labels
        return to_tfdataset(features=self, targets=targets, **kwargs)



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import logging
import os

from climetlab.readers import reader
from climetlab.utils import progress_bar, tqdm

LOG = logging.getLogger(__name__)


def _index_grib_file(path):
    import eccodes

    def parse_field(h):
        field = dict()

        field["_path"] = path

        i = eccodes.codes_keys_iterator_new(h, "mars")
        try:
            while eccodes.codes_keys_iterator_next(i):
                name = eccodes.codes_keys_iterator_get_name(i)
                value = eccodes.codes_get_string(h, name)
                field[name] = value

        finally:
            eccodes.codes_keys_iterator_delete(i)

        field["_offset"] = eccodes.codes_get_long(h, "offset")
        field["_length"] = eccodes.codes_get_long(h, "totalLength")

        field["_param_id"] = eccodes.codes_get_string(h, "paramId")
        field["param"] = eccodes.codes_get_string(h, "shortName")

        return field

    size = os.path.getsize(path)
    pbar = progress_bar(desc=f"Parsing {path}", total=size)

    with open(path, "rb") as f:
        old_position = f.tell()
        h = eccodes.codes_grib_new_from_file(f)

        while h:
            try:
                yield parse_field(h)
            finally:
                eccodes.codes_release(h)

            position = f.tell()
            pbar.update(position - old_position)
            old_position = position
            h = eccodes.codes_grib_new_from_file(f)
    pbar.close()


def _index_url(url):
    import climetlab as cml

    path = cml.load_source("url", url).path
    # TODO: should use download_and_cache
    # path = download_and_cache(url)
    for entry in _index_grib_file(path):
        del entry["_path"]
        yield entry


def _index_path(path):
    if os.path.isdir(path):
        for root, _, files in os.walk(path):
            for name in files:
                yield from _index_grib_file(os.path.join(root, name))
    else:
        yield from _index_grib_file(path)


class DirectoryParserIterator:
    """This class delays parsing the directory for the list of files
    until the iterator is actually used (calling __iter__)
    """

    def __init__(
        self, directory, relative_paths, ignore=None, followlinks=True, verbose=False
    ):
        if ignore is None:
            ignore = []
        self.ignore = ignore
        self.directory = directory
        self.relative_paths = relative_paths
        self.followlinks = followlinks
        self.verbose = verbose

        self._tasks = None

    def __iter__(self):
        for path in self.tasks:
            for entry in self.process_one_task(path):
                yield entry

    @property
    def tasks(self):
        if self._tasks is not None:
            return self._tasks

        LOG.debug(f"Parsing files in {self.directory}")
        assert os.path.exists(self.directory), f"{self.directory} does not exist"
        assert os.path.isdir(self.directory), f"{self.directory} is not a directory"

        tasks = []
        for root, _, files in os.walk(self.directory, followlinks=self.followlinks):
            for name in files:
                path = os.path.join(root, name)
                if path in self.ignore:
                    continue
                tasks.append(path)

        if tasks:
            if self.verbose:
                print(f"Found {len(tasks)} files to index.")
        else:
            LOG.error(f"Could not find any files to index in {self.directory}")

        self._tasks = tqdm(tasks)

        return self.tasks


class GribIndexingDirectoryParserIterator(DirectoryParserIterator):
    def process_one_task(self, path):
        LOG.debug(f"Parsing file {path}")

        if self.relative_paths is True:
            _path = os.path.relpath(path, self.directory)
        elif self.relative_paths is False:
            _path = os.path.abspath(path)
        elif self.relative_paths is None:
            _path = path
        else:
            assert False, self.relative_paths

        try:
            r = reader(self, path)
        except PermissionError as e:
            LOG.error(f"Could not read {path}: {e}")
            return

        for field in r.index_content():
            field["_path"] = _path
            yield field
